@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* exec.c: command execution */
-/* (C) 2007-2013 magicant */
+/* (C) 2007-2016 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -38,9 +38,7 @@
 #include <sys/times.h>
 #include <unistd.h>
 #include <wchar.h>
-#if YASH_ENABLE_ALIAS
-# include "alias.h"
-#endif
+#include "alias.h"
 #include "builtin.h"
 #include "expand.h"
 #if YASH_ENABLE_HISTORY
@@ -329,7 +327,7 @@ void exec_pipelines_async(const pipeline_T *p)
     
     if (cpid > 0) {
 	/* parent process: add a new job */
-	job_T *job = xmalloc(sizeof *job + sizeof *job->j_procs);
+	job_T *job = xmalloc(add(sizeof *job, sizeof *job->j_procs));
 	process_T *ps = job->j_procs;
 
 	ps->pr_pid = cpid;
@@ -434,8 +432,10 @@ void exec_for(const command_T *c, bool finally_exit)
     int i;
     for (i = 0; i < count; i++) {
 	if (!set_variable(c->c_forname, words[i],
-		    posixly_correct ? SCOPE_GLOBAL : SCOPE_LOCAL, false))
+		    posixly_correct ? SCOPE_GLOBAL : SCOPE_LOCAL, false)) {
+	    laststatus = Exit_ASSGNERR;
 	    goto done;
+	}
 	exec_and_or_lists(c->c_forcmds, finally_exit && i + 1 == count);
 
 	if (c->c_forcmds == NULL)
@@ -498,14 +498,14 @@ void exec_case(const command_T *c, bool finally_exit)
 {
     assert(c->c_type == CT_CASE);
 
-    wchar_t *word = expand_single(c->c_casword, TT_SINGLE);
+    wchar_t *word = expand_single_and_unescape(
+	    c->c_casword, TT_SINGLE, true, false);
     if (word == NULL)
 	goto fail;
-    word = unescapefree(word);
 
     for (const caseitem_T *ci = c->c_casitems; ci != NULL; ci = ci->next) {
 	for (void **pats = ci->ci_patterns; *pats != NULL; pats++) {
-	    wchar_t *pattern = expand_single(*pats, TT_SINGLE);
+	    wchar_t *pattern = expand_single(*pats, TT_SINGLE, true, false);
 	    if (pattern == NULL)
 		goto fail;
 
@@ -545,10 +545,9 @@ void exec_funcdef(const command_T *c, bool finally_exit)
 {
     assert(c->c_type == CT_FUNCDEF);
 
-    wchar_t *funcname = expand_single(c->c_funcname, TT_SINGLE);
+    wchar_t *funcname = expand_single_and_unescape(
+	    c->c_funcname, TT_SINGLE, true, false);
     if (funcname != NULL) {
-	funcname = unescapefree(funcname);
-
 	if (define_function(funcname, c->c_funcbody))
 	    laststatus = Exit_SUCCESS;
 	else
@@ -770,6 +769,11 @@ pid_t exec_process(
 	finally_exit = true;
     }
 
+    lastcmdsubstatus = Exit_SUCCESS;
+
+    /* connect pipes and close leftovers */
+    connect_pipes(pi);
+
     if (c->c_type == CT_SIMPLE) {
 	if (!expand_line(c->c_words, &argc, &argv)) {
 	    laststatus = Exit_EXPERROR;
@@ -789,11 +793,6 @@ pid_t exec_process(
     }
     /* `argc' and `argv' are used only for `CT_SIMPLE'. */
 
-    lastcmdsubstatus = Exit_SUCCESS;
-
-    /* connect pipes and close leftovers */
-    connect_pipes(pi);
-
     /* open redirections */
     savefd_T *savefd;
     if (!open_redirections(c->c_redirs, &savefd)) {
@@ -816,10 +815,13 @@ pid_t exec_process(
     last_assign = c->c_assigns;
     if (argc == 0) {
 	/* if there is no command word, just perform assignments */
-	if (do_assignments(c->c_assigns, false, shopt_allexport))
+	if (do_assignments(c->c_assigns, false, shopt_allexport)) {
 	    laststatus = lastcmdsubstatus;
-	else
+	} else {
 	    laststatus = Exit_ASSGNERR;
+	    if (!is_interactive_now)
+		finally_exit = true;
+	}
 	print_xtrace(NULL);
 	goto done2;
     }
@@ -985,7 +987,7 @@ pid_t fork_and_reset(pid_t pgid, bool fg, sigtype_T sigtype)
 	if (sigtype & t_leave) {
 	    clear_exit_trap();
 	} else {
-	    clear_traps();
+	    phantomize_traps();
 	    neglect_all_jobs();
 #if YASH_ENABLE_HISTORY
 	    close_history_file();
@@ -1361,8 +1363,7 @@ void exec_function_body(
     (void) complete;
 #endif
     exec_commands(body, finally_exit ? E_SELF : E_NORMAL);
-    if (execstate.exception == E_RETURN)
-	execstate.exception = E_NONE;
+    cancel_return();
     close_current_environment();
 
     execstate.noreturn = save_noreturn;
@@ -1931,6 +1932,7 @@ int dot_builtin(int argc, void **argv)
     }
 
     exec_input(fd, mbsfilename, enable_alias ? XIO_SUBST_ALIAS : 0);
+    cancel_return();
     remove_shellfd(fd);
     xclose(fd);
     free(mbsfilename);
@@ -1943,7 +1945,7 @@ int dot_builtin(int argc, void **argv)
 
 error:
     free(mbsfilename);
-    if (!is_interactive_now)
+    if (special_builtin_executed && !is_interactive_now)
 	exit_shell_with_status(Exit_FAILURE);
     return Exit_FAILURE;
 }
@@ -2076,7 +2078,7 @@ int exec_builtin_2(int argc, void **argv, const wchar_t *as, bool clear)
     }
 
     exec_external_program(commandpath, argc, mbsargv0, argv, envs);
-    err = Exit_NOEXEC;
+    err = laststatus;
 
     if (clear)
 	plfree((void **) envs, free);
@@ -2257,7 +2259,6 @@ bool print_command_info(
 	return true;
     }
 
-#if YASH_ENABLE_ALIAS
     if (aliases) {
 	if (print_alias_if_defined(commandname, humanfriendly)) {
 	    return true;
@@ -2266,9 +2267,6 @@ bool print_command_info(
 		return false;
 	}
     }
-#else
-    (void) aliases;
-#endif
 
     name = malloc_wcstombs(commandname);
     if (name == NULL)
