@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* option.c: option settings */
-/* (C) 2007-2012 magicant */
+/* (C) 2007-2016 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -109,6 +109,10 @@ bool shopt_verbose = false;
  * Corresponds to the -x/--xtrace option. */
 bool shopt_xtrace = false;
 
+/* If set, the "xtrace" option is not ignored while executing auxiliary
+ * commands. */
+bool shopt_traceall = true;
+
 /* If set, lines that start with a space are not saved in the history.
  * Corresponds to the --histspace option. */
 bool shopt_histspace = false;
@@ -144,6 +148,10 @@ bool shopt_nullglob = false;
  * Corresponds to the --braceexpand option. */
 bool shopt_braceexpand = false;
 
+/* If set, empty last field is preserved in field splitting.
+ * Corresponds to the --emptylastfield option. */
+bool shopt_emptylastfield = false;
+
 /* If set, it is allowed to overwrite existing files by redirections.
  * Corresponds to the +C/--clobber option. */
 bool shopt_clobber = true;
@@ -195,6 +203,7 @@ static const struct option_T shell_options[] = {
 #if YASH_ENABLE_LINEEDIT
     { 0,    0,    L"emacs",          &shopt_emacs,          true, },
 #endif
+    { 0,    0,    L"emptylastfield", &shopt_emptylastfield, true, },
     { L'e', 0,    L"errexit",        &shopt_errexit,        true, },
     { 0,    L'n', L"exec",           &shopt_exec,           true, },
     { 0,    0,    L"extendedglob",   &shopt_extendedglob,   true, },
@@ -222,6 +231,7 @@ static const struct option_T shell_options[] = {
     { 0,    0,    L"nullglob",       &shopt_nullglob,       true, },
     { 0,    0,    L"posixlycorrect", &posixly_correct,      true, },
     { L's', 0,    L"stdin",          &shopt_stdin,          false, },
+    { 0,    0,    L"traceall",       &shopt_traceall,       true, },
     { 0,    L'u', L"unset",          &shopt_unset,          true, },
     { L'v', 0,    L"verbose",        &shopt_verbose,        true, },
 #if YASH_ENABLE_LINEEDIT
@@ -251,7 +261,7 @@ enum normal_option_index_T {
 static const struct xgetopt_T normal_options[] = {
 #if YASH_ENABLE_HELP
     [NOI_HELP]      = { L'-', L"help",      OPTARG_NONE,     false,
-			&normal_options, },
+			(void *) &normal_options, },
 #endif
     [NOI_VERSION]   = { L'V', L"version",   OPTARG_NONE,     false, NULL, },
     [NOI_NOPROFILE] = { L'-', L"noprofile", OPTARG_NONE,     false, NULL, },
@@ -275,6 +285,9 @@ static int parse_option_character(
 static int parse_long_option(void *const *argv, bool enable,
 	struct shell_invocation_T *shell_invocation)
     __attribute__((nonnull(1)));
+static size_t collect_matching_shell_options(
+	const wchar_t *optstr, plist_T *options)
+    __attribute__((nonnull));
 static void search_shell_options(const wchar_t *optname, plist_T *resultlist)
     __attribute__((nonnull));
 static void search_normal_options(const wchar_t *optname, plist_T *resultlist,
@@ -294,6 +307,10 @@ static void update_le_convmeta_option(void);
 static int set_normal_option(const struct xgetopt_T *opt, const wchar_t *arg,
 	struct shell_invocation_T *shell_invocation)
     __attribute__((nonnull(1)));
+#if YASH_ENABLE_TEST
+static struct option_T *find_option_unique(const wchar_t *s)
+    __attribute__((nonnull,pure));
+#endif
 
 /* Parses the specified command line arguments and accordingly sets the shell
  * options and positional parameters.
@@ -386,15 +403,10 @@ int parse_short_option(void *const *argv, bool enable,
 	    }
 
 	    plist_T options;
-	    size_t shelloptindex;
-
-	    wchar_t *normoptname = normalize_option_name(optname);
 	    pl_init(&options);
-	    search_shell_options(normoptname, &options);
-	    shelloptindex = options.length;
-	    if (matchwcsprefix(normoptname, L"no"))
-		search_shell_options(&normoptname[2], &options);
-	    free(normoptname);
+
+	    size_t shelloptindex =
+		collect_matching_shell_options(optname, &options);
 
 	    return handle_search_result(&options, argv, enable,
 		    shelloptindex, options.length, shell_invocation);
@@ -447,33 +459,42 @@ int parse_long_option(void *const *argv, bool enable,
 
     pl_init(&options);
 
-    /* No long options are available in the POSIXly correct mode. */
     if (posixly_correct) {
+	/* No long options are available in the POSIXly correct mode. */
 	shelloptindex = noshelloptindex = 0;
-	goto end;
+    } else {
+	shelloptindex = collect_matching_shell_options(optstr, &options);
+	noshelloptindex = options.length;
+
+	if (enable) {
+	    assert(matchwcsprefix(optstr, L"--"));
+	    search_normal_options(&optstr[2], &options, shell_invocation);
+	}
     }
 
-    wchar_t *optname = normalize_option_name(optstr);
-
-    search_shell_options(optname, &options);
-    shelloptindex = options.length;
-
-    if (matchwcsprefix(optname, L"no"))
-	search_shell_options(&optname[2], &options);
-    noshelloptindex = options.length;
-
-    free(optname);
-
-    if (enable) {
-	assert(matchwcsprefix(optstr, L"--"));
-	search_normal_options(&optstr[2], &options, shell_invocation);
-    }
-
-end:
     return handle_search_result(&options, argv, enable,
 	    shelloptindex, noshelloptindex, shell_invocation);
 }
 
+/* Collects shell options that match the specified name.
+ * This function adds to `options' pointers to shell options (struct option_T)
+ * whose names start with `optname'.
+ * If `optname' starts with "no", options whose names start with `optname'
+ * without "no" are also added. The index of the first such option is returned.
+ */
+size_t collect_matching_shell_options(const wchar_t *optname, plist_T *options)
+{
+    wchar_t *normoptname = normalize_option_name(optname);
+    search_shell_options(normoptname, options);
+
+    size_t nooptindex = options->length;
+    if (matchwcsprefix(normoptname, L"no"))
+	search_shell_options(&normoptname[2], options);
+
+    free(normoptname);
+    return nooptindex;
+}
+
 /* Adds pointers to shell options (struct option_T) whose names match `optname'
  * to the specified list.
  * If `optname' exactly matches one of the shell option names, `resultlist' is
@@ -750,6 +771,41 @@ void set_lineedit_option(enum shopt_lineedit_T v)
 
 #endif
 
+#if YASH_ENABLE_TEST
+
+/* Returns true iff the specified string is a valid option name that can be used
+ * as the argument to the "-o" option. */
+bool is_valid_option_name(const wchar_t *s)
+{
+    return find_option_unique(s) != NULL;
+}
+
+/* Returns true iff the specified string is a valid option name that can be used
+ * as the argument to the "-o" option and the option is enabled. */
+bool option_is_enabled(const wchar_t *s)
+{
+    struct option_T *opt = find_option_unique(s);
+    return opt != NULL && *opt->optp;
+}
+
+/* If the specified string is a valid option name that can be used as the
+ * argument to the "-o" option, then returns the option. Otherwise, NULL is
+ * returned. */
+struct option_T *find_option_unique(const wchar_t *s)
+{
+    plist_T options;
+    pl_init(&options);
+
+    collect_matching_shell_options(s, &options);
+
+    struct option_T *opt = (options.length == 1) ? options.contents[0] : NULL;
+
+    pl_destroy(&options);
+    return opt;
+}
+
+#endif /* YASH_ENABLE_TEST */
+
 /* Returns the current value of special parameter $- as a newly malloced string.
  */
 wchar_t *get_hyphen_parameter(void)
