@@ -27,6 +27,7 @@
 #if HAVE_GETTEXT
 # include <libintl.h>
 #endif
+#include <limits.h>
 #if HAVE_PATHS_H
 # include <paths.h>
 #endif
@@ -218,7 +219,7 @@ bool is_same_file(const char *path1, const char *path2)
  * error. */
 wchar_t *canonicalize_path(const wchar_t *path)
 {
-    wchar_t *const result = xmallocn(wcslen(path) + 1, sizeof *result);
+    wchar_t *const result = xmalloce(wcslen(path), 1, sizeof *result);
     wchar_t *rp = result;
     plist_T clist;
 
@@ -315,13 +316,13 @@ char *xgetcwd(void)
 {
 #if GETCWD_AUTO_MALLOC
     char *pwd = getcwd(NULL, 0);
-    return (pwd != NULL) ? xrealloc(pwd, strlen(pwd) + 1) : NULL;
+    return (pwd != NULL) ? xrealloc(pwd, add(strlen(pwd), 1)) : NULL;
 #else
     size_t pwdlen = 40;
     char *pwd = xmalloc(pwdlen);
     while (getcwd(pwd, pwdlen) == NULL) {
 	if (errno == ERANGE) {
-	    pwdlen *= 2;
+	    pwdlen = mul(pwdlen, 2);
 	    pwd = xrealloc(pwd, pwdlen);
 	} else {
 	    int saveerrno = errno;
@@ -388,13 +389,16 @@ char *which(
 
 /* Creates a new temporary file under "/tmp".
  * `mode' is the access permission bits of the file.
+ * The name of the temporary file ends with `suffix`, which should be as short
+ * as possible and must not exceed 8 bytes.
  * On successful completion, a file descriptor is returned, which is both
  * readable and writeable regardless of `mode', and a pointer to a string
  * containing the filename is assigned to `*filename', which should be freed by
  * the caller. The filename consists only of portable filename characters.
  * On failure, -1 is returned with `**filename' left unchanged and `errno' is
  * set to the error value. */
-int create_temporary_file(char **filename, mode_t mode)
+int create_temporary_file(
+	char **restrict filename, const char *restrict suffix, mode_t mode)
 {
     static uintmax_t num = 0;
     uintmax_t n;
@@ -407,8 +411,13 @@ int create_temporary_file(char **filename, mode_t mode)
     sb_init(&buf);
     for (int i = 0; i < 100; i++) {
 	num = (num ^ n) * 16777619;
-	sb_printf(&buf, "/tmp/yash-%X", (unsigned) (num >> 30));
-	/* The filename must be 14 bytes long at most. */
+	sb_printf(&buf, "/tmp/yash-%" PRIXMAX, num);
+
+	size_t maxlen = _POSIX_NAME_MAX + 5 - strlen(suffix);
+	if (buf.length > maxlen)
+	    sb_truncate(&buf, maxlen);
+	sb_cat(&buf, suffix);
+
 	fd = open(buf.contents, O_RDWR | O_CREAT | O_EXCL, mode);
 	if (fd >= 0) {
 	    *filename = sb_tostr(&buf);
@@ -438,7 +447,10 @@ static wchar_t *get_default_path(void)
  * Keys are pointers to a multibyte string containing a command name and
  * values are pointers to a multibyte string containing the commands' full path.
  * For each entry, the key string is part of the value, that is, the last
- * pathname component of the value. */
+ * pathname component of the value.
+ * Full paths may be relative, in which case the paths are unreliable because
+ * the working directory may have been changed since the paths had been
+ * entered. */
 static hashtable_T cmdhash;
 
 /* Initializes the command hashtable. */
@@ -465,12 +477,12 @@ const char *get_command_path(const char *name, bool forcelookup)
 
     if (!forcelookup) {
 	path = ht_get(&cmdhash, name).value;
-	if (path != NULL && is_executable_regular(path))
+	if (path != NULL && path[0] == '/' && is_executable_regular(path))
 	    return path;
     }
 
     path = which(name, get_path_array(PA_PATH), is_executable_regular);
-    if (path != NULL && path[0] == '/') {
+    if (path != NULL) {
 	size_t namelen = strlen(name), pathlen = strlen(path);
 	const char *nameinpath = path + pathlen - namelen;
 	assert(strcmp(name, nameinpath) == 0);
@@ -1193,11 +1205,6 @@ int change_directory(const wchar_t *newpwd, bool printnewdir, bool logical)
 step6:  /* set the value of `curpath' */
     assert(newpwd[0] != L'/');
     assert(curpath.length == 0);
-    if (logical) {
-	wb_cat(&curpath, origpwd);
-	if (curpath.length == 0 || curpath.contents[curpath.length - 1] != L'/')
-	    wb_wccat(&curpath, L'/');
-    }
     wb_cat(&curpath, newpwd);
 
 step7:  /* ensure the value of `curpath' is an absolute path */
@@ -1482,8 +1489,11 @@ void print_command_paths(bool all)
     size_t index = 0;
 
     while ((kv = ht_next(&cmdhash, &index)).key != NULL) {
+	const char *path = kv.value;
+	if (path[0] != '/')
+	    continue;
 	if (all || get_builtin(kv.key) == NULL) {
-	    if (!xprintf("%s\n", (char *) kv.value)) {
+	    if (!xprintf("%s\n", path)) {
 		break;
 	    }
 	}
